<?php declare(strict_types=1);

namespace Addons\Integrations\Algolia\Controllers;

use Addons\Integrations\Algolia\Requests\SettingsRequest;
use App\Modules\Core\Controllers\Integrations\Integration;
use App\Modules\Core\Models\Integration as IntegrationModel;
use App\Modules\Core\Models\IntegrationService;
use Illuminate\Contracts\View\View;
use Illuminate\Http\RedirectResponse;
use Illuminate\Support\Facades\Artisan;
use Illuminate\Support\Facades\Config;
use JsValidator;
use Lang;
use Redirect;
use Session;
use TemplateView;

use function collect;
use function config;
use function is_numeric;

class Algolia extends Integration
{
    /**
     * Integration identifier.
     *
     * @var string
     */
    const IDENTIFIER = 'Algolia';

    /**
     * Integration settings route.
     *
     * @var string
     */
    const SETTINGS_ROUTE = 'integration.algolia.settings';

    /**
     * Initialise the integration.
     */
    public function __construct()
    {
        parent::__construct();

        $this->setIdentifier(self::IDENTIFIER);

        // Register the settings page.
        $this->registerSetting(self::SETTINGS_ROUTE);

        $this->replaceScoutDriver();
    }

    /**
     * Get the settings page.
     */
    public function getSettingsPage(): View
    {
        return TemplateView::operator($this->getNamespace() . '::settings')
            ->with('docsLink', 'https://docs.supportpal.com/current/Third+Party+Integrations+Algolia')
            ->with('jsValidator', JsValidator::formRequest(SettingsRequest::class))
            ->with('settings', $this->getSettings());
    }

    /**
     * Update the settings.
     */
    public function updateSettings(SettingsRequest $request): RedirectResponse
    {
        $data = $request->all(['app_id', 'app_secret']);

        // Work through each row of data.
        foreach ($data as $key => $value) {
            if (! empty($value) || is_numeric($value)) {
                $this->addSetting($key, $value);
            } else {
                $this->removeSetting($key);
            }
        }

        // Add scheduled task to index models.
        Artisan::call('models:index');

        Session::flash('success', Lang::get('messages.success_settings'));

        return Redirect::route(self::SETTINGS_ROUTE);
    }

    /**
     * Plugins can run an installation routine when they are activated. This will typically include adding default
     * values, initialising database tables and so on.
     *
     * @return boolean
     */
    public function activate()
    {
        // Register the services that the integration can be used for.
        $this->registerService(IntegrationService::SERVICE_SEARCH);

        return true;
    }

    /**
     * Deactivating serves as temporarily disabling the plugin, but the files still remain. This function should
     * typically clear any caches and temporary directories.
     *
     * @return boolean
     */
    public function deactivate()
    {
        return true;
    }

    /**
     * When a plugin is uninstalled, it should be completely removed as if it never was there. This function should
     * delete any created database tables, and any files created outside of the plugin directory.
     *
     * @return boolean
     */
    public function uninstall()
    {
        $this->removeSettings();

        return true;
    }

    private function replaceScoutDriver(): void
    {
        /** @var IntegrationModel|null $integration */
        $integration = Config::get('integration')->where('id', $this->getModel('id'))->first();
        if ($integration === null) {
            return;
        }

        /** @var array<string,string> $settings */
        $settings = $integration->settings;

        // Not configured.
        if (collect($settings)->isEmpty() || (empty($settings['app_id']) && empty($settings['app_secret']))) {
            return;
        }

        // Update configuration.
        config()->set('scout.algolia.id', $settings['app_id']);
        config()->set('scout.algolia.secret', $settings['app_secret']);
        config()->set('scout.driver', 'algolia');
    }
}
