<?php declare(strict_types=1);

namespace Addons\Reports\Feedback\Reports\Cards;

use Addons\Reports\Feedback\Controllers\ModelConcerns;
use App\Http\Requests\Request;
use App\Modules\Core\Models\FeedbackFormFieldValue;
use App\Modules\Core\Models\FeedbackLog;
use App\Modules\Core\Models\FeedbackRating;
use App\Modules\Report\Addon\Dashboards\Cards\HtmlCard;
use App\Modules\Report\Addon\Dashboards\Cards\Number\NumberFormat;
use App\Modules\Report\Addon\Dashboards\Cards\UnderlyingData;
use App\Modules\Report\Addon\Dashboards\Filtering\QueryBuilder;
use App\Modules\Selfservice\Models\Article;
use App\Modules\Ticket\Models\Ticket;
use Illuminate\Contracts\View\View;
use Illuminate\Database\Eloquent\Builder;
use TemplateView;

use function get_class;

/**
 * @template TModel of Ticket|Article
 * @template TFeedbackLog of FeedbackLog
 * @template TFeedbackFormFieldValue of FeedbackFormFieldValue
 * @implements UnderlyingData<TModel>
 */
abstract class RatingCard extends HtmlCard implements UnderlyingData
{
    /** @phpstan-use ModelConcerns<TModel, TFeedbackLog, TFeedbackFormFieldValue> */
    use ModelConcerns;

    protected int $width = 4;
    protected int $height = 2;

    abstract public function getFeedbackRating(): FeedbackRating;

    public function html(): View
    {
        $given = $this->totalGivenRatings();
        $total = $this->totalRatings();

        return TemplateView::make('Reports#Feedback::rating')
            ->with('icon', $this->getFeedbackRating()->faIcon())
            ->with('colour', $this->getFeedbackRating()->textColour())
            ->with('percentage', $this->toPercent($given, $total))
            ->with('count', $given)
            ->with('total', $total);
    }

    /**
     * @inheritDoc
     */
    public function underlyingDataQuery(Request $request): Builder
    {
        return $this->baseQuery()
            ->with('feedback')
            ->where($this->getQualifiedName($this->logModel()->getTable(), 'rating'), '=', $this->getFeedbackRating()->value)
            ->selectRaw($this->getQualifiedName($this->model()->getTable(), '*'));
    }

    private function toPercent(int $given, int $total): string
    {
        $number = $total === 0 ? 0 : $given / $total;

        return NumberFormat::PERCENT->format($number);
    }

    private function totalRatings(): int
    {
        return $this->baseQuery()->count();
    }

    private function totalGivenRatings(): int
    {
        return $this->baseQuery()
            ->where($this->getQualifiedName($this->logModel()->getTable(), 'rating'), '=', $this->getFeedbackRating()->value)
            ->groupBy($this->getQualifiedName($this->logModel()->getTable(), 'rating'))
            ->count();
    }

    /**
     * @return Builder<TModel>
     */
    private function baseQuery(): Builder
    {
        /** @var class-string<TModel> $model */
        $model = get_class($this->model());
        $filters = $this->dashboard()->filters($this->getQualifiedName($this->logModel()->getTable(), 'updated_at'));

        return QueryBuilder::for($model)
            ->applyFilters($filters)
            ->join(
                $this->logModel()->getTable(),
                $this->logRelation()->getQualifiedParentKeyName(),
                '=',
                $this->logRelation()->getQualifiedForeignKeyName()
            )
            ->whereNotNull($this->getQualifiedName($this->logModel()->getTable(), 'rating'));
    }
}
