<?php declare(strict_types=1);

namespace Addons\Reports\Feedback\Reports\Cards;

use Addons\Reports\Feedback\Controllers\ModelConcerns;
use App\Http\Requests\Request;
use App\Modules\Core\Models\FeedbackFormFieldValue;
use App\Modules\Core\Models\FeedbackLog;
use App\Modules\Core\Models\FeedbackRating;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\ChartType;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\Datasets\PrimitiveDataset;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\DataStructure;
use App\Modules\Report\Addon\Dashboards\Cards\ChartCard;
use App\Modules\Report\Addon\Dashboards\Cards\UnderlyingData;
use App\Modules\Report\Addon\Dashboards\Filtering\QueryBuilder;
use App\Modules\Selfservice\Models\Article;
use App\Modules\Ticket\Models\Ticket;
use Illuminate\Database\Eloquent\Builder;
use UnhandledMatchError;

use function get_class;
use function sprintf;
use function trans;

/**
 * @template TModel of Ticket|Article
 * @template TFeedbackLog of FeedbackLog
 * @template TFeedbackFormFieldValue of FeedbackFormFieldValue
 * @implements UnderlyingData<TModel>
 */
abstract class ResponseRateChart extends ChartCard implements UnderlyingData
{
    /** @phpstan-use ModelConcerns<TModel, TFeedbackLog, TFeedbackFormFieldValue> */
    use ModelConcerns;

    protected int $width = 6;

    public function name(): string
    {
        return trans('report.feedback_response_rate');
    }

    public function data(): DataStructure
    {
        $results = $this->baseQuery()
            ->groupBy('rating')
            ->selectRaw(sprintf(
                'IF(ISNULL(%s), "null", %s) as rating, count(*) as aggregate',
                $this->getQualifiedName($this->logModel()->getTable(), 'rating'),
                $this->getQualifiedName($this->logModel()->getTable(), 'rating')
            ))
            ->getQuery()
            ->pluck('aggregate', 'rating')
            ->toArray();

        $unanswered = $results['null'] ?? 0;
        $answered = ($results[FeedbackRating::BAD->value] ?? 0)
            + ($results[FeedbackRating::NEUTRAL->value] ?? 0)
            + ($results[FeedbackRating::GOOD->value] ?? 0);

        return (new DataStructure(ChartType::PIE, new PrimitiveDataset([$answered, $unanswered])))
            ->setLabels([trans('report.feedback_answered'), trans('report.feedback_unanswered')]);
    }

    /**
     * @inheritDoc
     */
    public function underlyingDataQuery(Request $request): Builder
    {
        $query = $this->baseQuery();
        if ($this->isDrillIntoChartRequest($request)) {
            $label = $request->string('label')->toString();

            if ($label === trans('report.feedback_answered')) {
                $query->whereNotNull('rating');
            } elseif ($label === trans('report.feedback_unanswered')) {
                $query->whereNull('rating');
            } else {
                throw new UnhandledMatchError(sprintf('Unsupported label \'%s\'.', $label));
            }
        }

        return $query->with('feedback')
            ->selectRaw($this->getQualifiedName($this->model()->getTable(), '*'));
    }

    /**
     * @return Builder<TModel>
     */
    private function baseQuery(): Builder
    {
        /** @var class-string<TModel> $model */
        $model = get_class($this->model());
        $filters = $this->dashboard()->filters($this->getQualifiedName($this->logModel()->getTable(), 'updated_at'));

        return QueryBuilder::for($model)
            ->applyFilters($filters)
            ->join(
                $this->logModel()->getTable(),
                $this->logRelation()->getQualifiedParentKeyName(),
                '=',
                $this->logRelation()->getQualifiedForeignKeyName()
            );
    }

    private function isDrillIntoChartRequest(Request $request): bool
    {
        return $request->has('label');
    }
}
