<?php declare(strict_types=1);

namespace Addons\Reports\Feedback\Reports\Cards\Responses;

use App\Http\Requests\Request;
use App\Modules\Core\Models\FeedbackFormFieldOption;
use App\Modules\Core\Models\FeedbackFormFieldValue;
use App\Modules\Core\Models\FeedbackLog;
use App\Modules\Core\Models\TranslationCollection;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\ChartType;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\Datasets\PrimitiveDataset;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\DataStructure;
use App\Modules\Selfservice\Models\Article;
use App\Modules\Ticket\Models\Ticket;
use Illuminate\Database\Eloquent\Builder;
use UnhandledMatchError;

use function array_column;
use function array_sum;
use function count;
use function escape_like;
use function json_decode;
use function sprintf;

/**
 * @template TModel of Ticket|Article
 * @template TFeedbackLog of FeedbackLog
 * @template TFeedbackFormFieldValue of FeedbackFormFieldValue
 * @template TFeedbackFormFieldOption of FeedbackFormFieldOption
 * @extends FieldChartCard<TModel, TFeedbackLog, TFeedbackFormFieldValue>
 */
abstract class RatingFieldChart extends FieldChartCard
{
    public function data(): DataStructure
    {
        $results = $this->getChartJsDataset();

        return (new DataStructure(ChartType::BAR, new PrimitiveDataset(array_column($results, 'rating'))))
            ->setLabels(array_column($results, 'label'))
            ->setOptions([
                'indexAxis' => 'y',
                'plugins' => [
                    'legend' => ['display' => false]
                ],
                'scales' => [
                    'x' => [
                        'max' => 5
                    ],
                ],
            ]);
    }

    /**
     * @inheritDoc
     */
    public function underlyingDataQuery(Request $request): Builder
    {
        $query = $this->baseQuery();
        if ($this->isDrillIntoChartRequest($request)) {
            $matched = false;
            $label = $request->string('label')->toString();

            foreach ($this->getFieldOptions() as $option) {
                if ($label !== $option->value) {
                    continue;
                }

                $query->where('value', 'LIKE', '%"' . escape_like((string) $option->id) . '":%');

                $matched = true;
                break;
            }

            if (! $matched) {
                throw new UnhandledMatchError(sprintf('Unsupported label \'%s\'.', $label));
            }
        }

        return $query->with('feedback')
            ->selectRaw($this->getQualifiedName($this->model()->getTable(), '*'));
    }

    /**
     * @return TranslationCollection<int, TFeedbackFormFieldOption>
     */
    abstract protected function getFieldOptions(): TranslationCollection;

    /**
     * @return mixed[]
     */
    private function getChartJsDataset(): array
    {
        $options = $this->getFieldOptions();

        $results = [];
        foreach ($options as $option) {
            $results[$option->id] = [
                'label' => $option->value,
                'ratings' => []
            ];
        }

        $values = $this->baseQuery()
            ->select('value')
            ->getQuery()
            ->get();

        foreach ($values as $value) {
            $value = json_decode($value->value);

            foreach ((array) $value as $key => $rating) {
                $results[$key]['ratings'][] = $rating;
            }
        }

        foreach ($results as $key => $result) {
            $results[$key]['rating'] = empty($result['ratings']) ? 0 : array_sum($result['ratings']) / count($result['ratings']);
        }

        return $results;
    }
}
