<?php declare(strict_types=1);

namespace Addons\Reports\Feedback\Reports\Cards;

use Addons\Reports\Feedback\Controllers\ModelConcerns;
use App\Http\Requests\Request;
use App\Modules\Core\Models\FeedbackFormFieldValue;
use App\Modules\Core\Models\FeedbackLog;
use App\Modules\Report\Addon\Concerns\InteractsWithTimeframe;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\ChartType;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\Datasets\ObjectDataset;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\DataStructure;
use App\Modules\Report\Addon\Dashboards\Cards\ChartCard;
use App\Modules\Report\Addon\Dashboards\Cards\UnderlyingData;
use App\Modules\Report\Addon\Dashboards\Filtering\QueryBuilder;
use App\Modules\Selfservice\Models\Article;
use App\Modules\Ticket\Models\Ticket;
use Carbon\CarbonInterface;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Support\Collection;
use Illuminate\Support\ItemNotFoundException;
use UnhandledMatchError;

use function get_class;
use function get_object_vars;
use function sprintf;
use function trans;

/**
 * @template TModel of Ticket|Article
 * @template TFeedbackLog of FeedbackLog
 * @template TFeedbackFormFieldValue of FeedbackFormFieldValue
 * @implements UnderlyingData<TModel>
 */
abstract class TrendChart extends ChartCard implements UnderlyingData
{
    /** @phpstan-use ModelConcerns<TModel, TFeedbackLog, TFeedbackFormFieldValue> */
    use ModelConcerns, InteractsWithTimeframe;

    protected int $width = 6;

    public function name(): string
    {
        return trans('report.feedback_trend');
    }

    public function data(): DataStructure
    {
        $results = $this->getChartJsDatasets();

        $data = $results->map(fn($item) => ['x' => $item['x'], 'y' => $item['good']])->toArray();
        $dataset = new ObjectDataset($data, $this->goodLabel());
        $dataset->setOptions(['backgroundColor' => '#34d399']);

        $dataStructure = new DataStructure(ChartType::BAR, $dataset);

        $data = $results->map(fn($item) => ['x' => $item['x'], 'y' => $item['neutral']])->toArray();
        $dataset = new ObjectDataset($data, $this->neutralLabel());
        $dataStructure->addDataset($dataset->setOptions(['backgroundColor' => '#fcd34d']));

        $data = $results->map(fn($item) => ['x' => $item['x'], 'y' => $item['bad']])->toArray();
        $dataset = new ObjectDataset($data, $this->badLabel());
        $dataStructure->addDataset($dataset->setOptions(['backgroundColor' => '#f87171']));

        return $dataStructure
            ->setAxisNames($this->unitToTrans($this->calculateUnits()), $this->yAxisName())
            ->setOptions($this->options());
    }

    /**
     * @inheritDoc
     */
    public function underlyingDataQuery(Request $request): Builder
    {
        $query = $this->baseQuery();

        if ($this->isDrillIntoChartRequest($request)) {
            $rating = $this->labelToRating($request->string('label')->toString());
            $time = $request->integer('value.x') / 1000;

            $query->where($this->getQualifiedName($this->logModel()->getTable(), 'rating'), $rating)
                ->whereRaw(sprintf('%s = ?', $this->getTimeSeriesExpr($this->timeframeColumn())), $time);
        }

        return $query->with('feedback')
            ->selectRaw($this->getQualifiedName($this->model()->getTable(), '*'));
    }

    abstract public function yAxisName(): string;

    private function goodLabel(): string
    {
        return trans('general.good');
    }

    private function neutralLabel(): string
    {
        return trans('general.neutral');
    }

    private function badLabel(): string
    {
        return trans('general.bad');
    }

    private function isDrillIntoChartRequest(Request $request): bool
    {
        return $request->has('label') && $request->has('value.x') && $request->has('value.y');
    }

    /**
     * @return QueryBuilder<TModel>
     */
    private function baseQuery(): QueryBuilder
    {
        /** @var class-string<TModel> $model */
        $model = get_class($this->model());
        $filters = $this->dashboard()->filters($this->timeframeColumn());

        return QueryBuilder::for($model)
            ->applyFilters($filters)
            ->join(
                $this->logModel()->getTable(),
                $this->logRelation()->getQualifiedParentKeyName(),
                '=',
                $this->logRelation()->getQualifiedForeignKeyName()
            )
            ->whereNotNull($this->getQualifiedName($this->logModel()->getTable(), 'rating'));
    }

    /**
     * @return Collection<int,mixed>
     */
    private function getChartJsDatasets(): Collection
    {
        $collection = $this->baseQuery()
            ->selectRaw(sprintf('%s * 1000 as x', $this->getTimeSeriesExpr($this->timeframeColumn())))
            ->selectRaw('COUNT(rating = 0 or null) as bad')
            ->selectRaw('COUNT(rating = 1 or null) as good')
            ->selectRaw('COUNT(rating = 2 or null) as neutral')
            ->groupBy('x')
            ->getQuery()
            ->get()
            ->keyBy('x');

        return $this->createDateRange($this->calculateUnits())
            ->map(function (CarbonInterface $d) use ($collection) {
                if (isset($collection[$d->unix() * 1000])) {
                    return get_object_vars($collection[$d->unix() * 1000]);
                }

                return ['x' => $d->unix() * 1000, 'bad' => 0, 'good' => 0, 'neutral' => 0];
            });
    }

    private function timeframeColumn(): string
    {
        return $this->getQualifiedName($this->logModel()->getTable(), 'updated_at');
    }

    /**
     * @throws UnhandledMatchError|ItemNotFoundException
     */
    private function labelToRating(string $label): int
    {
        return match ($label) {
            $this->goodLabel()    => 1,
            $this->neutralLabel() => 2,
            $this->badLabel()     => 0,
            default               => throw new UnhandledMatchError(sprintf('Unsupported label \'%s\'.', $label)),
        };
    }

    /**
     * @return mixed[]
     */
    private function options(): array
    {
        $unit = $this->calculateUnits();

        return [
            'scales' => [
                'x' => [
                    'stacked' => true,
                    'type'    => 'time',
                    'time'    => [
                        'unit' => $unit,
                        'round' => $unit,
                        'tooltipFormat' => $this->unitToTooltipFormat($unit),
                    ],
                ],
                'y' => [
                    'min'     => 0,
                    'stacked' => true,
                    'ticks'   => ['precision' => 0],
                ],
            ]
        ];
    }
}
