<?php declare(strict_types=1);

namespace Addons\Reports\TicketPerformance\Reports\Cards\DepartmentPerformance;

use App\Http\Requests\Request;
use App\Modules\Report\Addon\Dashboards\Cards\Table\Column;
use App\Modules\Report\Addon\Dashboards\Cards\TableCard;
use App\Modules\Report\Addon\Dashboards\Cards\UnderlyingData;
use App\Modules\Report\Addon\Dashboards\Filtering\QueryBuilder;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketCreatedAtColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketDepartmentColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketNumberColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketSubjectColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\UserFormattedNameColumn;
use App\Modules\Report\Metrics\Metric;
use App\Modules\Ticket\Models\Ticket;
use Carbon\CarbonInterval;
use Illuminate\Database\Eloquent\Builder;

use function e;
use function sprintf;
use function trans;
use function trans_choice;

/**
 * @phpstan-extends TableCard<Ticket>
 * @implements UnderlyingData<Ticket>
 */
class FirstResponseTimeTable extends TableCard implements UnderlyingData
{
    protected int $width = 6;

    public function name(): string
    {
        return trans('report.first_response_time_average');
    }

    /**
     * @return QueryBuilder<Ticket>
     */
    public function query(): QueryBuilder
    {
        return $this->baseQuery()
            ->groupBy('ticket.department_id')
            ->select('ticket.department_id', 'department.name')
            ->selectRaw('AVG(report_metric_value.value) as `first_response_time`');
    }

    public function columns(): array
    {
        return [
            $this->departmentColumn(),
            $this->timeColumn(),
        ];
    }

    protected function departmentColumn(): Column
    {
        return (new Column(trans_choice('ticket.department', 1)))
            ->setValue(fn($model) => $model->name);
    }

    protected function timeColumn(): Column
    {
        return (new Column(trans('report.response_time')))
            ->setValue(fn($model) => CarbonInterval::seconds($model->first_response_time)->cascade()->forHumans(null, true, 3));
    }

    public function getUnderlyingDataValueKey(): string
    {
        return 'department_id';
    }

    /**
     * @return Builder<Ticket>
     */
    public function underlyingDataQuery(Request $request): Builder
    {
        $query = $this->baseQuery()
            ->with('department', 'user')
            ->selectRaw('ticket.*')
            ->selectRaw('CAST(report_metric_value.value as unsigned) as `first_response_time`');

        if ($this->isDrillIntoChartRequest($request)) {
            $query->where('ticket.department_id', $request->integer('value'));
        }

        return $query;
    }

    public function underlyingDataTableColumns(): array
    {
        $timeColumn = (new Column(trans('report.response_time')))
            ->setId('first_response_time')
            ->setDefinition(['width' => '160px'])
            ->setCsvValue(fn($ticket) => CarbonInterval::seconds($ticket->first_response_time)->cascade()->forHumans())
            ->setValue(function ($ticket) {
                $carbon = CarbonInterval::seconds($ticket->first_response_time)->cascade();
                $fullTime = $carbon->forHumans();
                $formattedTime = $carbon->forHumans(null, true, 3);

                return sprintf('<time title="%s">%s</time>', e($fullTime), e($formattedTime));
            }, false);

        return [
            $timeColumn,
            new TicketCreatedAtColumn,
            new TicketNumberColumn,
            new TicketSubjectColumn,
            new UserFormattedNameColumn,
            new TicketDepartmentColumn,
        ];
    }

    /**
     * @return QueryBuilder<Ticket>
     */
    private function baseQuery(): QueryBuilder
    {
        return QueryBuilder::for(Ticket::class)
            ->applyFilters($this->dashboard()->filters('ticket.created_at'))
            ->join('department', 'ticket.department_id', '=', 'department.id')
            ->join('report_metric_value', 'report_metric_value.model_id', '=', 'ticket.id')
            ->join('report_metric', 'report_metric_value.metric_id', '=', 'report_metric.id')
            ->where('report_metric.name', '=', Metric::FIRST_RESPONSE_TIME->value);
    }

    protected function isDrillIntoChartRequest(Request $request): bool
    {
        return $request->has('value');
    }
}
