<?php declare(strict_types=1);

namespace Addons\Reports\TicketPerformance\Reports\Cards\FirstResponseTime;

use App\Http\Requests\Request;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\ChartType;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\Datasets\PrimitiveDataset;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\DataStructure;
use App\Modules\Report\Addon\Dashboards\Filtering\QueryBuilder;
use App\Modules\Report\Metrics\Metric;
use App\Modules\Ticket\Models\Ticket;
use Illuminate\Database\Eloquent\Builder;
use UnhandledMatchError;

use function array_values;
use function sprintf;
use function trans;
use function trans_choice;

class ByTimeChart extends BaseChart
{
    public function name(): string
    {
        return trans('report.first_response_time');
    }

    public function data(): DataStructure
    {
        return (new DataStructure(ChartType::BAR, new PrimitiveDataset($this->getChartJsDataset())))
            ->setAxisNames(trans_choice('ticket.ticket', 2), trans('general.time'))
            ->setLabels([
                trans('report.0-15_mins'),
                trans('report.15-30_mins'),
                trans('report.30-60_mins'),
                trans('report.1-3_hours'),
                trans('report.3-6_hours'),
                trans('report.6-12_hours'),
                trans('report.12-24_hours'),
                trans('report.1-3_days'),
                trans('report.3_plus_days'),
            ])
            ->setOptions([
                'indexAxis' => 'y',
                'plugins' => [
                    'legend' => ['display' => false]
                ],
                'scales' => [
                    'x' => [
                        'ticks' => ['precision' => 0],
                        'title' => ['display' => true, 'text' => trans_choice('ticket.ticket', 2)],
                    ],
                ],
            ]);
    }

    /**
     * @return Builder<Ticket>
     */
    public function underlyingDataQuery(Request $request): Builder
    {
        $query = $this->baseQuery()
            ->with(['department', 'user'])
            ->select(['ticket.*'])
            ->selectRaw('CAST(report_metric_value.value as unsigned) as `first_response_time`');

        if ($this->isDrillIntoChartRequest($request)) {
            $label = $request->string('label')->toString();

            $exp = match ($label) {
                trans('report.0-15_mins')   => $this->lt15MinsExpr(),
                trans('report.15-30_mins')  => $this->fifteenTo30MinsExpr(),
                trans('report.30-60_mins')  => $this->thirtyTo60MinsExpr(),
                trans('report.1-3_hours')   => $this->oneTo3HoursExpr(),
                trans('report.3-6_hours')    => $this->threeTo6HoursExp(),
                trans('report.6-12_hours')   => $this->sixTo12HoursExp(),
                trans('report.12-24_hours')  => $this->twelveTo24HoursExp(),
                trans('report.1-3_days')    => $this->oneToThreeDaysExp(),
                trans('report.3_plus_days') => $this->gtThreeDaysExp(),
                default => throw new UnhandledMatchError
            };

            $query->whereRaw($exp);
        }

        return $query;
    }

    /**
     * @return int[]
     */
    private function getChartJsDataset(): array
    {
        $results = $this->baseQuery()
            ->selectRaw(sprintf('SUM(case when %s then 1 else 0 end) as `0`', $this->lt15MinsExpr()))
            ->selectRaw(sprintf('SUM(case when %s then 1 else 0 end) as `1`', $this->fifteenTo30MinsExpr()))
            ->selectRaw(sprintf('SUM(case when %s then 1 else 0 end) as `2`', $this->thirtyTo60MinsExpr()))
            ->selectRaw(sprintf('SUM(case when %s then 1 else 0 end) as `3`', $this->oneTo3HoursExpr()))
            ->selectRaw(sprintf('SUM(case when %s then 1 else 0 end) as `4`', $this->threeTo6HoursExp()))
            ->selectRaw(sprintf('SUM(case when %s then 1 else 0 end) as `5`', $this->sixTo12HoursExp()))
            ->selectRaw(sprintf('SUM(case when %s then 1 else 0 end) as `6`', $this->twelveTo24HoursExp()))
            ->selectRaw(sprintf('SUM(case when %s then 1 else 0 end) as `7`', $this->oneToThreeDaysExp()))
            ->selectRaw(sprintf('SUM(case when %s then 1 else 0 end) as `8`', $this->gtThreeDaysExp()))
            ->getQuery()
            ->get()
            ->firstOrFail();

        return array_values((array) $results);
    }

    private function lt15MinsExpr(): string
    {
        return 'FLOOR((report_metric_value.value) / 60) < 15';
    }

    private function fifteenTo30MinsExpr(): string
    {
        return 'FLOOR((report_metric_value.value) / 60) >= 15 and FLOOR((report_metric_value.value) / 60) < 30';
    }

    private function thirtyTo60MinsExpr(): string
    {
        return 'FLOOR((report_metric_value.value) / 60) >= 30 and FLOOR((report_metric_value.value) / 3600) < 1';
    }

    private function oneTo3HoursExpr(): string
    {
        return 'FLOOR((report_metric_value.value) / 3600) >= 1 and FLOOR((report_metric_value.value) / 3600) < 3';
    }

    private function threeTo6HoursExp(): string
    {
        return 'FLOOR((report_metric_value.value) / 3600) >= 3 and FLOOR((report_metric_value.value) / 3600) < 6';
    }

    private function sixTo12HoursExp(): string
    {
        return 'FLOOR((report_metric_value.value) / 3600) >= 6 and FLOOR((report_metric_value.value) / 3600) < 12';
    }

    private function twelveTo24HoursExp(): string
    {
        return 'FLOOR((report_metric_value.value) / 3600) >= 12 and FLOOR((report_metric_value.value) / 3600) < 24';
    }

    private function oneToThreeDaysExp(): string
    {
        return 'FLOOR((report_metric_value.value) / 3600) >= 24 and FLOOR((report_metric_value.value) / 3600) < 96';
    }

    private function gtThreeDaysExp(): string
    {
        return 'FLOOR((report_metric_value.value) / 3600) >= 96';
    }

    /**
     * @return QueryBuilder<Ticket>
     */
    private function baseQuery(): QueryBuilder
    {
        return QueryBuilder::for(Ticket::class)
            ->applyFilters($this->dashboard()->filters('ticket.created_at'))
            ->join('report_metric_value', 'report_metric_value.model_id', '=', 'ticket.id')
            ->join('report_metric', 'report_metric_value.metric_id', '=', 'report_metric.id')
            ->where('report_metric.name', '=', Metric::FIRST_RESPONSE_TIME->value);
    }
}
