<?php declare(strict_types=1);

namespace Addons\Reports\TicketPerformance\Reports\Cards\TimeToResolution;

use App\Http\Requests\Request;
use App\Modules\Report\Addon\Concerns\InteractsWithTimeframe;
use App\Modules\Report\Addon\Concerns\UnixTimestamp;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\ChartType;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\Datasets\ObjectDataset;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\DataStructure;
use App\Modules\Report\Addon\Dashboards\Cards\ChartCard;
use App\Modules\Report\Addon\Dashboards\Cards\Table\Column;
use App\Modules\Report\Addon\Dashboards\Cards\UnderlyingData;
use App\Modules\Report\Addon\Dashboards\Filtering\QueryBuilder;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketCreatedAtColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketDepartmentColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketNumberColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketSubjectColumn;
use App\Modules\Report\Metrics\Metric;
use App\Modules\Ticket\Models\Ticket;
use Carbon\CarbonInterface;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\Config;

use function get_object_vars;
use function sprintf;
use function trans;
use function trans_choice;

/**
 * @implements UnderlyingData<Ticket>
 */
class TrendChart extends ChartCard implements UnderlyingData
{
    use InteractsWithTimeframe, UnixTimestamp;

    protected int $width = 8;

    public function name(): string
    {
        return trans('report.resolution_time_trend');
    }

    public function data(): DataStructure
    {
        $unit = $this->calculateUnits();
        $data = $this->getChartJsDataset()->toArray();

        return (new DataStructure(ChartType::LINE, new ObjectDataset($data)))
            ->setAxisNames($this->unitToTrans($unit), trans_choice('general.hour', 2))
            ->setOptions([
                'plugins' => [
                    'legend' => ['display' => false]
                ],
                'scales' => [
                    'x' => [
                        'type'    => 'time',
                        'time'    => [
                            'unit' => $unit,
                            'round' => $unit,
                            'tooltipFormat' => $this->unitToTooltipFormat($unit),
                        ],
                    ],
                    'y' => ['min' => 0,],
                ]
            ]);
    }

    public function underlyingDataTableColumns(): array
    {
        $timeColumn = (new Column(trans('report.time_to_resolution')))
            ->setId('time_to_resolution')
            ->setDefinition(['width' => '160px'])
            ->setValue(fn($ticket) => $this->formatTime((int) $ticket->time_to_resolution));

        return [
            $timeColumn,
            new TicketCreatedAtColumn,
            new TicketNumberColumn,
            new TicketSubjectColumn,
            new TicketDepartmentColumn,
        ];
    }

    /**
     * @return Builder<Ticket>
     */
    public function underlyingDataQuery(Request $request): Builder
    {
        $query = $this->baseQuery()
            ->with('department')
            ->select(['ticket.*'])
            ->selectRaw('CAST(report_metric_value.value as UNSIGNED) as `time_to_resolution`');

        if ($this->isDrillIntoChartRequest($request)) {
            $time = $request->integer('value.x') / 1000;

            $query->whereRaw(sprintf('%s = ?', $this->getTimeSeriesExpr($this->timeframeColumn())), $time);
        }

        return $query;
    }

    /**
     * @return Collection<int,mixed>
     */
    private function getChartJsDataset(): Collection
    {
        $collection = $this->baseQuery()
            ->selectRaw(sprintf('%s * 1000 as x', $this->getTimeSeriesExpr($this->timeframeColumn())))
            ->selectRaw('AVG(report_metric_value.value) / 3600.0 as y')
            ->groupBy('x')
            ->getQuery()
            ->get()
            ->keyBy('x');

        return $this->createDateRange($this->calculateUnits())
            ->map(function (CarbonInterface $d) use ($collection) {
                if (isset($collection[$d->unix() * 1000])) {
                    return get_object_vars($collection[$d->unix() * 1000]);
                }

                return ['x' => $d->unix() * 1000, 'y' => 0];
            });
    }

    /**
     * @return QueryBuilder<Ticket>
     */
    private function baseQuery(): QueryBuilder
    {
        return QueryBuilder::for(Ticket::class)
            ->applyFilters($this->dashboard()->filters($this->timeframeColumn()))
            ->join('report_metric_value', 'report_metric_value.model_id', '=', 'ticket.id')
            ->join('report_metric', 'report_metric_value.metric_id', '=', 'report_metric.id')
            ->where('ticket.status_id', Config::get('settings.default_resolved_status'))
            ->where('report_metric.name', '=', Metric::TIME_TO_RESOLUTION->value);
    }

    private function timeframeColumn(): string
    {
        return 'ticket.resolved_time';
    }

    private function isDrillIntoChartRequest(Request $request): bool
    {
        return $request->has('label') && $request->has('value.x') && $request->has('value.y');
    }
}
