<?php declare(strict_types=1);

namespace Addons\Reports\TicketVolume\Reports\Cards\TicketsOpened;

use App\Http\Requests\Request;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\ChartType;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\Datasets\CustomObjectDataset;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\DataStructure;
use App\Modules\Report\Addon\Dashboards\Cards\ChartCard;
use App\Modules\Report\Addon\Dashboards\Cards\UnderlyingData;
use App\Modules\Report\Addon\Dashboards\Filtering\QueryBuilder;
use App\Modules\Ticket\Models\Ticket;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\DB;

/**
 * @implements UnderlyingData<Ticket>
 */
abstract class BaseChart extends ChartCard implements UnderlyingData
{
    protected int $width = 6;

    /**
     * @return class-string
     */
    abstract public function chartModel(): string;

    abstract public function queryColumn(): string;

    public function modelNameAttribute(): string
    {
        return 'name';
    }

    public function data(): DataStructure
    {
        $data = $this->getChartJsDataset();
        $dataset = new CustomObjectDataset($data->toArray(), 'aggregate', 'name');

        return (new DataStructure(ChartType::BAR, $dataset))
            ->setLabels($data->pluck('name')->toArray())
            ->setOptions([
                'indexAxis' => 'y',
                'plugins' => [
                    'legend' => ['display' => false]
                ],
                'scales' => [
                    'x' => [
                        'ticks' => ['precision' => 0]
                    ],
                ],
            ]);
    }

    /**
     * @inheritDoc
     */
    public function underlyingDataQuery(Request $request): Builder
    {
        $query = $this->baseQuery();

        if ($this->isDrillIntoChartRequest($request)) {
            $query->where($this->queryColumn(), $request->integer('value.id'));
        }

        return $query->with($this->underlyingDataRelations())
            ->selectRaw('ticket.*');
    }

    /**
     * @return string[]
     */
    protected function underlyingDataRelations(): array
    {
        return ['user', 'department', 'status'];
    }

    /**
     * @return Builder<Ticket>
     */
    protected function baseQuery(): Builder
    {
        return QueryBuilder::for(Ticket::class)
            ->applyFilters($this->dashboard()->filters('ticket.created_at'));
    }

    /**
     * @return Collection<int,int>
     */
    protected function getAggregates(string $column): Collection
    {
        return $this->baseQuery()
            ->groupBy($column)
            ->select([$column, DB::raw('count(*) as aggregate')])
            ->pluck('aggregate', $column);
    }

    protected function isDrillIntoChartRequest(Request $request): bool
    {
        return $request->has('value.id');
    }

    /**
     * @return Collection<string, mixed>
     */
    private function getChartJsDataset(): Collection
    {
        $results = $this->getAggregates($this->queryColumn());

        return $this->chartModel()::all()
            ->map(function ($item) use ($results) {
                return [
                    'id'        => $item->id,
                    'name'      => $item->{$this->modelNameAttribute()},
                    'aggregate' => $results[$item->id] ?? 0,
                ];
            });
    }
}
