<?php declare(strict_types=1);

namespace Addons\Reports\UserStatistics\Reports\Cards\Activity;

use App\Http\Requests\Request;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\ChartType;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\Datasets\CustomObjectDataset;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\DataStructure;
use App\Modules\Report\Addon\Dashboards\Cards\ChartCard;
use App\Modules\Report\Addon\Dashboards\Cards\UnderlyingData;
use App\Modules\Report\Addon\Dashboards\Filtering\QueryBuilder;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketSubjectColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\UserOrganisationNameColumn;
use App\Modules\Ticket\Models\Ticket;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Support\Collection;
use UnhandledMatchError;

use function trans;
use function trans_choice;

/**
 * @implements UnderlyingData<Ticket>
 */
class OrganisationsWithMostTickets extends ChartCard implements UnderlyingData
{
    protected int $width = 6;

    public function name(): string
    {
        return trans('report.organisations_with_most_tickets');
    }

    public function data(): DataStructure
    {
        $results = $this->getChartJsDataset();
        $dataset = new CustomObjectDataset($results->toArray(), 'name', 'aggregate');

        return (new DataStructure(ChartType::BAR, $dataset))
            ->setLabels($results->pluck('name')->toArray())
            ->setAxisNames(trans_choice('user.organisation', 1), trans_choice('ticket.ticket', 2))
            ->setOptions([
                'plugins' => [
                    'legend' => ['display' => false]
                ],
                'scales' => [
                    'y' => [
                        'ticks' => ['precision' => 0],
                    ],
                ]
            ]);
    }

    /**
     * @inheritDoc
     */
    public function underlyingDataTableColumns(): array
    {
        return [
            new UserOrganisationNameColumn,
            new TicketSubjectColumn,
        ];
    }

    /**
     * @inheritDoc
     */
    public function underlyingDataQuery(Request $request): Builder
    {
        $dataset = $this->getChartJsDataset();
        if ($this->isDrillIntoChartRequest($request)) {
            // We could just pass the user_id straight into the DB, but it's best we
            // check the user_id actually exists in the data set from a permissions perspective...
            $id = $request->string('value.organisation_id')->toString();
            $datapoint = $dataset->keyBy('organisation_id')->get($id);
            if ($datapoint === null) {
                throw new UnhandledMatchError('The given user id does not exist in the data set.');
            }

            $ids = [$datapoint['organisation_id']];
        } else {
            $ids = $dataset->pluck('organisation_id')->toArray();
        }

        return QueryBuilder::for(Ticket::class)
            ->applyFilters($this->dashboard()->filters('ticket.created_at'))
            ->whereIn('user_organisation.id', $ids)
            ->with('user.organisation')
            ->join('user', 'ticket.user_id', '=', 'user.id')
            ->join('user_organisation', 'user.organisation_id', '=', 'user_organisation.id')
            ->select(['user_organisation.name', 'ticket.*']);
    }

    private function isDrillIntoChartRequest(Request $request): bool
    {
        return $request->has('value.organisation_id');
    }

    /**
     * @return Collection<int, array{'organisation_id': mixed, 'aggregate': mixed, 'name': mixed}>
     */
    private function getChartJsDataset(): Collection
    {
        return QueryBuilder::for(Ticket::class)
            ->applyFilters($this->dashboard()->filters('ticket.created_at'))
            ->join('user', 'ticket.user_id', '=', 'user.id')
            ->join('user_organisation', 'user.organisation_id', '=', 'user_organisation.id')
            ->where('ticket.internal', 0)
            ->whereNotNull('user.organisation_id')
            ->groupBy('user.organisation_id')
            ->orderBy('aggregate', 'desc')
            ->limit(5)
            ->selectRaw('user.organisation_id, user_organisation.name, count(ticket.id) as aggregate')
            ->get()
            // we map it because all sorts of other attributes are included in the results
            // and that then gets passed around between ChartJS and underlying data requests.
            ->map(function ($data) {
                return [
                    'organisation_id'   => $data['organisation_id'],
                    'aggregate'         => $data['aggregate'],
                    'name'              => $data['name'],
                ];
            });
    }
}
