<?php declare(strict_types=1);

namespace Addons\Reports\UserStatistics\Reports\Cards\Activity;

use App\Http\Requests\Request;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\ChartType;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\Datasets\CustomObjectDataset;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\DataStructure;
use App\Modules\Report\Addon\Dashboards\Cards\ChartCard;
use App\Modules\Report\Addon\Dashboards\Cards\UnderlyingData;
use App\Modules\Report\Addon\Dashboards\Filtering\QueryBuilder;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\TicketSubjectColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\Ticket\UserFormattedNameColumn;
use App\Modules\Ticket\Models\Ticket;
use App\Modules\User\Models\User;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Database\Eloquent\Relations\BelongsTo;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\DB;
use UnhandledMatchError;

use function trans;
use function trans_choice;

/**
 * @implements UnderlyingData<Ticket>
 */
class UsersWithMostTickets extends ChartCard implements UnderlyingData
{
    protected int $width = 6;

    public function name(): string
    {
        return trans('report.users_with_most_tickets');
    }

    public function data(): DataStructure
    {
        $results = $this->getChartJsDataset();
        $dataset = new CustomObjectDataset($results->toArray(), 'formatted_name', 'aggregate');

        return (new DataStructure(ChartType::BAR, $dataset))
            ->setLabels($results->pluck('formatted_name')->toArray())
            ->setAxisNames(trans_choice('user.user', 1), trans_choice('ticket.ticket', 2))
            ->setOptions([
                'plugins' => [
                    'legend' => ['display' => false]
                ],
                'scales' => [
                    'y' => [
                        'ticks' => ['precision' => 0],
                    ],
                ]
            ]);
    }

    /**
     * @inheritDoc
     */
    public function underlyingDataTableColumns(): array
    {
        return [
            new UserFormattedNameColumn,
            new TicketSubjectColumn,
        ];
    }

    /**
     * @inheritDoc
     */
    public function underlyingDataQuery(Request $request): Builder
    {
        $dataset = $this->getChartJsDataset();
        if ($this->isDrillIntoChartRequest($request)) {
            // We could just pass the user_id straight into the DB, but it's best we
            // check the user_id actually exists in the data set from a permissions perspective...
            $userId = $request->string('value.user_id')->toString();
            $datapoint = $dataset->keyBy('user_id')->get($userId);
            if ($datapoint === null) {
                throw new UnhandledMatchError('The given user id does not exist in the data set.');
            }

            $ids = [$datapoint['user_id']];
        } else {
            $ids = $dataset->pluck('user_id')->toArray();
        }

        return QueryBuilder::for(Ticket::class)
            ->applyFilters($this->dashboard()->filters('ticket.created_at'))
            ->whereIn('user_id', $ids)
            ->with('user')
            ->join('user', 'ticket.user_id', '=', 'user.id')
            ->select([DB::raw(User::getFormattedNameSql()), 'ticket.*']);
    }

    private function isDrillIntoChartRequest(Request $request): bool
    {
        return $request->has('value.user_id');
    }

    /**
     * @return Collection<int, array{'user_id': mixed, 'aggregate': mixed, 'formatted_name': mixed}>
     */
    private function getChartJsDataset(): Collection
    {
        return QueryBuilder::for(Ticket::class)
            ->applyFilters($this->dashboard()->filters('ticket.created_at'))
            ->with([
                'user' => function (BelongsTo $query) {
                    $query->select('id', 'firstname', 'lastname', 'email', 'role');
                }
            ])
            ->where('ticket.internal', 0)
            ->groupBy('ticket.user_id')
            ->orderBy('aggregate', 'desc')
            ->limit(5)
            ->selectRaw('ticket.*, count(ticket.user_id) as aggregate')
            ->get()
            // we map it because all sorts of other attributes are included in the results
            // and that then gets passed around between ChartJS and underlying data requests.
            ->map(function ($data) {
                return [
                    'user_id'   => $data['user_id'],
                    'aggregate' => $data['aggregate'],
                    'formatted_name' => $data['user']['formatted_name'],
                ];
            });
    }
}
