<?php declare(strict_types=1);

namespace Addons\Reports\UserStatistics\Reports\Cards\Demographics;

use App\Http\Requests\Request;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\ChartType;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\Datasets\ChoroplethDataset;
use App\Modules\Report\Addon\Dashboards\Cards\Chart\DataStructure;
use App\Modules\Report\Addon\Dashboards\Cards\ChartCard;
use App\Modules\Report\Addon\Dashboards\Cards\UnderlyingData;
use App\Modules\Report\Addon\Dashboards\Filtering\QueryBuilder;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\User\UserCountryColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\User\UserCreatedAtColumn;
use App\Modules\Report\Addon\Dashboards\UnderlyingData\Columns\User\UserFormattedNameColumn;
use App\Modules\User\Models\User;
use Illuminate\Database\Eloquent\Builder;
use Illuminate\Support\Collection;
use SupportPal\TopoClient\Feature;
use SupportPal\TopoClient\Locale\Locales;
use SupportPal\TopoClient\TopoClient;

use function in_array;
use function trans;

/**
 * @implements UnderlyingData<User>
 */
class CountryChart extends ChartCard implements UnderlyingData
{
    protected int $width = 12;
    protected int $height = 8;

    public function name(): string
    {
        return trans('report.users_by_country');
    }

    public function data(): DataStructure
    {
        $countries = $this->countries();

        $dataset = (new ChoroplethDataset($this->getChartJsDataset($countries)))
            ->setOptions(['outline' => $countries->values()->toArray()]);

        return (new DataStructure(ChartType::CHOROPLETH, $dataset))
            ->setLabels($countries->map(fn (Feature $f) => $f['properties']['name'])->values()->toArray())
            ->setOptions([
                'showOutline' => false,
                'showGraticule' => false,
                'plugins' => [
                    'legend' => ['display' => false],
                ],
                'scales' => [
                    'projection' => [
                        'axis' => 'x',
                        'projection' => 'mercator',
                    ],
                ],
            ]);
    }

    public function getUnderlyingDataValueKey(): string
    {
        return 'alpha2';
    }

    public function underlyingDataTableColumns(): array
    {
        return [
            new UserCreatedAtColumn,
            new UserFormattedNameColumn,
            new UserCountryColumn,
        ];
    }

    public function underlyingDataQuery(Request $request): Builder
    {
        $query = $this->baseQuery();
        if ($this->isDrillIntoChartRequest($request)) {
            $query->where('user.country', $request->string('value')->toString());
        }

        return $query->select(['user.*', User::getFormattedNameSql()]);
    }

    /**
     * @return Collection<int, Feature>
     */
    private function countries(): Collection
    {
        $exclude = [
            '010' // Antarctica
        ];

        return TopoClient::countries()
            ->filter(function (Feature $f) use ($exclude) {
                return $f->getId() !== null && ! in_array($f->getId(), $exclude);
            });
    }

    /**
     * @return QueryBuilder<User>
     */
    private function baseQuery(): QueryBuilder
    {
        return QueryBuilder::for(User::class)
            ->applyFilters($this->dashboard()->filters())
            ->user()
            ->whereNotNull('user.country');
    }

    private function isDrillIntoChartRequest(Request $request): bool
    {
        return $request->has('label');
    }

    /**
     * @param Collection<int, Feature> $countries
     * @return mixed[]
     */
    private function getChartJsDataset(Collection $countries): array
    {
        $locales = new Locales;
        $dataset = $this->baseQuery()
            ->selectRaw('user.country, COUNT(user.id) as aggregate')
            ->groupBy('user.country')
            ->getQuery()
            ->pluck('aggregate', 'country');

        return $countries
            ->map(function (Feature $d) use ($dataset, $locales) {
                return [
                    'alpha2'  => $alpha2 = $locales->findId((string) $d->getId())->alpha2(),
                    'feature' => $d->toArray(),
                    'value'   => $dataset[$alpha2] ?? 0,
                ];
            })
            ->values()
            ->toArray();
    }
}
