<?php declare(strict_types=1);

namespace SupportPal\Addons\Support;

use Closure;
use Illuminate\Contracts\Events\Dispatcher;

use function sprintf;

trait HasEvents
{
    /**
     * The event dispatcher instance.
     *
     * @var Dispatcher
     */
    protected static $eventDispatcher;

    /**
     * Remove all the event listeners for the model.
     */
    public static function flushEventListeners(): void
    {
        if (! isset(static::$eventDispatcher)) {
            return;
        }

        foreach (AddonEvent::cases() as $event) {
            static::$eventDispatcher->forget(self::eventName($event));
        }
    }

    /**
     * Get the event dispatcher instance.
     *
     * @return Dispatcher
     */
    public static function getEventDispatcher()
    {
        return static::$eventDispatcher;
    }

    /**
     * Set the event dispatcher instance.
     *
     * @param Dispatcher $dispatcher
     */
    public static function setEventDispatcher(Dispatcher $dispatcher): void
    {
        static::$eventDispatcher = $dispatcher;
    }

    /**
     * Unset the event dispatcher for models.
     */
    public static function unsetEventDispatcher(): void
    {
        static::$eventDispatcher = null;
    }

    /**
     * Register a model event with the dispatcher.
     *
     * @param  Closure|string|mixed[]  $callback
     */
    public function on(AddonEvent $event, $callback): void
    {
        if (! isset(static::$eventDispatcher)) {
            return;
        }

        static::$eventDispatcher->listen(static::eventName($event), $callback);
    }

    public function fire(AddonEvent $event, bool $halt = true): mixed
    {
        if (! isset(static::$eventDispatcher)) {
            return true;
        }

        // First, we will get the proper method to call on the event dispatcher, and then we
        // will attempt to fire a custom, object based event for the given event. If that
        // returns a result we can return that result, or we'll call the string events.
        $method = $halt ? 'until' : 'dispatch';

        return ! empty($result) ? $result : static::$eventDispatcher->{$method}(self::eventName($event), $this);
    }

    protected static function eventName(AddonEvent $event): string
    {
        return sprintf('addon.%s: %s', $event->value, static::class);
    }
}
