<?php declare(strict_types=1);

namespace Chumper\Datatable\Engines;

use Chumper\Datatable\Columns\BaseColumn;
use Chumper\Datatable\Columns\DateColumn;
use Chumper\Datatable\Columns\FunctionColumn;
use Chumper\Datatable\Columns\TextColumn;
use Chumper\Datatable\Engines\Dataset\Dataset;
use Exception;
use Illuminate\Contracts\Database\Query\Expression;
use Illuminate\Http\JsonResponse;
use Illuminate\Http\Request;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\Config;
use Illuminate\Support\Facades\Response;
use Illuminate\Support\Str;

use function array_keys;
use function array_merge;
use function array_search;
use function count;
use function func_get_args;
use function in_array;
use function intval;
use function is_array;
use function is_callable;
use function is_numeric;
use function method_exists;
use function str_replace;
use function strpos;
use function substr;

abstract class BaseEngine
{
    const ORDER_ASC = 'asc';
    const ORDER_DESC = 'desc';

    /** @var array */
    protected $config = array();

    /** @var mixed */
    protected $rowClass = null;

    /** @var mixed */
    protected $rowId = null;

    /** @var array */
    protected $rowData = null;

    protected array $columnSearches = array();

    /**
     * @var array
     * support for DB::raw fields on where
     */
    protected $fieldSearches = array();

    /**
     * @var array
     * support for DB::raw fields on where
     * sburkett - added for column-based exact matching
     */
    protected $columnSearchExact = array();

    /** @var */
    protected $sEcho;

    /** @var Collection<mixed, mixed> */
    protected Collection $columns;

    /** @var array */
    protected $searchColumns = array();

    /** @var array */
    protected $showColumns = array();

    /** @var array */
    protected $orderColumns = array();

    protected int $skip = 0;

    protected ?int $limit = null;

    /** @var null */
    protected $search = null;

    /**
     * @var null
     * Will be an array if order is set
     * array(
     *  0 => column
     * )
     */
    protected $orderColumn = null;

    /** @var string */
    protected $orderDirection = BaseEngine::ORDER_ASC;

    /** @var boolean If the return should be alias mapped */
    protected $aliasMapping = false;

    /** @var bool If the search should be done with exact matching */
    protected $exactWordSearch = false;

    /**
     * Flag to skip total records count query.
     *
     * @var bool
     */
    protected bool $skipTotalRecords = false;

    protected int $totalRecords = 0;
    protected int $filteredRecords = 0;

    public function __construct(private readonly Request $request)
    {
        $this->columns = new Collection;
        $this->config = Config::get('chumper.datatable.engine');
        $this->setExactWordSearch($this->config['exactWordSearch']);
    }

    public function addColumn(BaseColumn|string $column, callable|string $value = ''): static
    {
        if ($column instanceof BaseColumn) {
            $this->columns->put($column->getName(), $column);
        } elseif (is_callable($value)) {
            $this->columns->put($column, new FunctionColumn($column, $value));
        } else {
            $this->columns->put($column, new TextColumn($column, $value));
        }

        return $this;
    }

    public function getColumn(string $name): BaseColumn
    {
        return $this->columns->get($name);
    }

    /** @return Collection<mixed, mixed> */
    public function getColumns(): Collection
    {
        return $this->columns;
    }

    public function getOrder(): array
    {
        return array_keys($this->columns->toArray());
    }

    /**
     * @return array
     */
    public function getOrderingColumns()
    {
        return $this->orderColumns;
    }

    /**
     * @return array
     */
    public function getSearchingColumns()
    {
        return $this->searchColumns;
    }

    /**
     * @return $this
     */
    public function clearColumns()
    {
        $this->columns = new Collection;

        return $this;
    }

    /**
     * @param array|string $cols
     */
    public function showColumns($cols): static
    {
        if (! is_array($cols)) {
            $cols = func_get_args();
        }

        foreach ($cols as $property) {
            //quick fix for created_at and updated_at columns
            if (in_array($property, array('created_at', 'updated_at'))) {
                $this->columns->put($property, new DateColumn($property, DateColumn::DAY_DATE));
            } else {
                $this->columns->put($property, new FunctionColumn($property, function ($model) use ($property) {
                    try {
                        return is_array($model) ? $model[$property] : $model->$property;
                    } catch (Exception $e) {
                        return null;
                    }
                }));
            }

            $this->showColumns[] = $property;
        }

        return $this;
    }

    public function skipTotalRecords(): static
    {
        $this->skipTotalRecords = true;

        return $this;
    }

    public function dataset(): Dataset
    {
        $this->handleInputs();
        $this->prepareSearchColumns();

        $dataset = $this->configureDataset();

        if (! $this->skipTotalRecords) {
            $this->totalRecords = $this->countRecords();
        }

        $dataset = $this->filter($dataset);
        $dataset = $this->sort($dataset);

        return $this->paginate($dataset);
    }

    /**
     * Make JSON response.
     *
     * @param array $extra Extra data to add to the response. Cannot override default response data!
     * @return JsonResponse
     */
    public function make(array $extra = [])
    {
        $aaData = $this->dataset()->get();
        $length = $aaData->count();

        $output = array(
            'aaData' => $aaData,
            'sEcho' => intval($this->sEcho),
            'iTotalRecords' => $this->skipTotalRecords ? $length : $this->totalCount(),
            'iTotalDisplayRecords' => $this->skipTotalRecords ? $length : $this->count(),
        );

        return Response::json(array_merge($extra, $output), 200, [ 'Content-Type' => 'application/json; charset=utf-8' ]);
    }

    /**
     * @param array|string|Expression $cols
     * @return $this
     */
    public function searchColumns($cols): static
    {
        if (! is_array($cols)) {
            $cols = func_get_args();
        }

        $this->searchColumns = $cols;

        return $this;
    }

    /**
     * @param array|string $cols
     */
    public function orderColumns($cols): static
    {
        if (! is_array($cols)) {
            $cols = func_get_args();
        }

        if (count($cols) === 1 && $cols[0] === '*') {
            $cols = $this->showColumns;
        }

        $this->orderColumns = $cols;

        return $this;
    }

    /**
     * @param  callable $function Set a function for a dynamic row class
     * @return $this
     */
    public function setRowClass($function): static
    {
        $this->rowClass = $function;

        return $this;
    }

    /**
     * @param  callable $function Set a function for a dynamic row id
     * @return $this
     */
    public function setRowId($function)
    {
        $this->rowId = $function;

        return $this;
    }

    /**
     * @param  callable $function Set a function for dynamic html5 data attributes
     * @return $this
     */
    public function setRowData($function): static
    {
        $this->rowData = $function;

        return $this;
    }

    public function setAliasMapping(bool $value = true): static
    {
        $this->aliasMapping = $value;

        return $this;
    }

    public function setExactWordSearch(bool $value = true): static
    {
        $this->exactWordSearch = $value;

        return $this;
    }

    /**
     * @param  array $columnNames Sets up a lookup table for which columns should use exact matching -sburkett
     * @return $this
     */
    public function setExactMatchColumns($columnNames): static
    {
        foreach ($columnNames as $columnIndex) {
            $this->columnSearchExact[ $columnIndex ] = true;
        }

        return $this;
    }

    public function getRowClass()
    {
        return $this->rowClass;
    }

    public function getRowId()
    {
        return $this->rowId;
    }

    public function getRowData()
    {
        return $this->rowData;
    }

    public function getAliasMapping()
    {
        return $this->aliasMapping;
    }

    protected function handleiDisplayStart(mixed $value): void
    {
        if (! is_numeric($value)) {
            return;
        }

        $this->skip((int) $value);
    }

    protected function handleiDisplayLength(mixed $value): void
    {
        if (! is_numeric($value)) {
            return;
        }

        $this->take((int) $value);
    }

    protected function handlesEcho(mixed $value): void
    {
        $this->sEcho = $value;
    }

    protected function handlesSearch(mixed $value): void
    {
        $this->search($value);
    }

    // phpcs:ignore PSR1.Methods.CamelCapsMethodName.NotCamelCaps
    protected function handleiSortCol_0(mixed $value): void
    {
        $direction = $this->request->input('sSortDir_0') === 'desc' ? BaseEngine::ORDER_DESC : BaseEngine::ORDER_ASC;

        //check if order is allowed
        if (empty($this->orderColumns)) {
            $this->order(array(0 => $value, 1 => $this->getNameByIndex($value)), $direction);

            return;
        }

        //prepare order array
        $cleanNames = array();
        foreach ($this->orderColumns as $c) {
            if (strpos($c, ':') !== false) {
                $cleanNames[] = substr($c, 0, strpos($c, ':'));
            } else {
                $cleanNames[] = $c;
            }
        }

        $i = 0;
        foreach ($this->columns as $name => $column) {
            if ($i === (int) $value && in_array($name, $cleanNames)) {
                $this->order(array(0 => $value, 1 => $this->orderColumns[array_search($name, $cleanNames)]), $direction);

                return;
            }

            $i++;
        }
    }

    /**
     * @param array-key $columnIndex
     * @param string $searchValue
     */
    protected function handleSingleColumnSearch($columnIndex, $searchValue): void
    {
        if (! isset($this->searchColumns[$columnIndex])
            || (empty($searchValue) && $searchValue !== '0')
        ) {
            return;
        }

        $columnName = $this->searchColumns[$columnIndex];
        $this->searchOnColumn($columnName, $searchValue);
    }

    protected function handleInputs(): void
    {
        foreach ($this->request->all() as $key => $input) {
            if ($this->isParameterForSingleColumnSearch($key)) {
                $columnIndex = str_replace('sSearch_', '', $key);
                $this->handleSingleColumnSearch($columnIndex, $input);
                continue;
            }

            if (! method_exists($this, $function = 'handle'.$key)) {
                continue;
            }

            $this->$function($input);
        }
    }

    /**
     * @param string $parameterName
     */
    protected function isParameterForSingleColumnSearch($parameterName): bool
    {
        static $parameterNamePrefix = 'sSearch_';

        return Str::contains($parameterName, $parameterNamePrefix);
    }

    protected function prepareSearchColumns(): void
    {
        if (count($this->searchColumns) !== 0 && ! empty($this->searchColumns)) {
            return;
        }

        $this->searchColumns = $this->showColumns;
    }

    /**
     * @param array $column
     * @param string $order
     */
    protected function order($column, $order = BaseEngine::ORDER_ASC): void
    {
        $this->orderColumn = $column;
        $this->orderDirection = $order;
    }

    protected function search(mixed $value): void
    {
        $this->search = $value;
    }

    /**
     * @param string $columnName
     * @param mixed $value
     */
    protected function searchOnColumn($columnName, $value)
    {
        $this->fieldSearches[] = $columnName;
        $this->columnSearches[] = $value;
    }

    protected function skip(int $value): void
    {
        $this->skip = $value;
    }

    protected function take(int $value): void
    {
        $this->limit = $value;
    }

    /**
     * @param int $index
     */
    public function getNameByIndex($index): mixed
    {
        $i = 0;
        foreach ($this->columns as $name => $col) {
            if ((int) $index === $i) {
                return $name;
            }

            $i++;
        }

        return null;
    }

    public function getExactWordSearch(): bool
    {
        return $this->exactWordSearch;
    }

    protected function totalCount(): int
    {
        return $this->totalRecords;
    }

    protected function count(): int
    {
        return $this->filteredRecords;
    }

    protected function filter(Dataset $dataset): Dataset
    {
        return $dataset;
    }

    protected function sort(Dataset $dataset): Dataset
    {
        return $dataset;
    }

    protected function paginate(Dataset $dataset): Dataset
    {
        return $dataset;
    }

    abstract protected function configureDataset(): Dataset;

    abstract protected function countRecords(): int;
}
