<?php declare(strict_types=1);

namespace Chumper\Datatable\Engines;

use Chumper\Datatable\Database\Counter;
use Chumper\Datatable\Engines\Dataset\Dataset;
use Chumper\Datatable\Engines\Dataset\QueryDataset;
use Illuminate\Contracts\Database\Query\Builder;
use Illuminate\Http\Request;

class QueryEngine extends BaseEngine
{
    /**
     * Flag to ignore the selects in count query.
     *
     * @var bool
     */
    protected bool $ignoreSelectInCountQuery = false;

    /**
     * Flag to keep the select bindings.
     *
     * @var bool
     */
    protected bool $keepSelectBindings = false;

    public function __construct(protected Builder $query, Request $request)
    {
        parent::__construct($request);
    }

    /**
     * Ignore the selects in count query.
     *
     * @return $this
     */
    public function ignoreSelectsInCountQuery(): static
    {
        $this->ignoreSelectInCountQuery = true;

        return $this;
    }

    /**
     * Keep the select bindings.
     *
     * @return $this
     */
    public function keepSelectBindings(): static
    {
        $this->keepSelectBindings = true;

        return $this;
    }

    protected function countRecords(): int
    {
        return (new Counter($this->query))
            ->ignoreSelectsInCountQuery($this->ignoreSelectInCountQuery)
            ->keepSelectBindings($this->keepSelectBindings)
            ->count();
    }

    protected function configureDataset(): Dataset
    {
        return new QueryDataset($this->query, $this);
    }

    protected function filter(Dataset $dataset): Dataset
    {
        $initialQuery = clone $this->query;

        if (! empty($this->search)) {
            $this->buildSearchQuery();
        }

        if (! empty($this->columnSearches)) {
            $this->buildSingleColumnSearches();
        }

        // If no modification between the original query and the filtered one has been made the filteredRecords equals
        // the totalRecords
        // Needs to be a loose comparison. The identity operator (===), object variables are identical if and only
        // if they refer to the same instance of the same class...
        if ($this->query == $initialQuery) { // phpcs:ignore SlevomatCodingStandard.Operators.DisallowEqualOperators.DisallowedEqualOperator
            $this->filteredRecords = $this->totalRecords;
        } elseif (! $this->skipTotalRecords) {
            $this->filteredRecords = $this->countRecords();
        }

        return $dataset;
    }

    protected function sort(Dataset $dataset): Dataset
    {
        if ($this->orderColumn !== null) {
            $i = 0;

            foreach ($this->columns as $col) {
                if ($i === (int) $this->orderColumn[0]) {
                    $this->query->orderBy($col->getName(), $this->orderDirection);
                    break;
                }

                $i++;
            }
        }

        return $dataset;
    }

    protected function paginate(Dataset $dataset): Dataset
    {
        if ($this->skip > 0) {
            $this->query->skip($this->skip);
        }

        if ($this->limit > 0) {
            $this->query->take($this->limit);
        }

        return $dataset;
    }

    private function buildSearchQuery(): void
    {
        $this->query->where(function ($query) {
            foreach ($this->searchColumns as $c) {
                $query->orWhere(
                    $c,
                    $this->exactWordSearch ? '=' : 'like',
                    $this->exactWordSearch ? $this->search : '%' . $this->search . '%'
                );
            }
        });
    }

    private function buildSingleColumnSearches(): void
    {
        foreach ($this->columnSearches as $index => $searchValue) {
            if (@ $this->columnSearchExact[$this->fieldSearches[$index]] === true) {
                $this->query->where($this->fieldSearches[$index], '=', $searchValue);
            } else {
                $this->query->where($this->fieldSearches[$index], 'like', '%' . $searchValue . '%');
            }
        }
    }
}
